#ifndef ASPEN_DDCOMPLEX_H_INCLUDED

#include "aspen_ddreal.h"

struct __align__(16) __cuddcomplex__
{
  cuddreal x, y;
};

typedef	struct __cuddcomplex__	cuddcomplex;


#ifdef __cplusplus

#ifdef __CUDA_ARCH__
#if GPU_ARCH>300
__device__ __forceinline__ cuddcomplex
__ldg ( const cuddcomplex * a )
{
  const cuddreal tx = __ldg( &(a->x) );
  const cuddreal ty = __ldg( &(a->y) );
  const cuddcomplex t = { tx, ty };
  return t;
}
#endif
#endif



__host__ __device__ __forceinline__ bool
operator== ( const cuddcomplex a, const cuddcomplex b )
{
  const bool ret = ((a.x == b.x) && (a.y == b.y)) ? true : false;
  return ret;
}

__host__ __device__ __forceinline__ bool
operator!= ( const cuddcomplex a, const cuddcomplex b )
{
  const bool ret = ((a.x != b.x) || (a.y != b.y)) ? true : false;
  return ret;
}


__host__ __device__ __forceinline__ cuddcomplex
operator+ ( const cuddcomplex a )
{
  return (a);
}

__host__ __device__ __forceinline__ cuddcomplex
operator+ ( const cuddcomplex a, const cuddcomplex b )
{
  cuddcomplex t;
  add2 ( a.x, b.x, t.x,  a.y, b.y, t.y );
  return t;
}

__host__ __device__ __forceinline__ void
operator+= ( cuddcomplex &a, const cuddcomplex b )
{
  a = (a + b);
}

__host__ __device__ __forceinline__ cuddcomplex
operator- ( const cuddcomplex a )
{
  const cuddreal tx = -a.x;
  const cuddreal ty = -a.y;
  const cuddcomplex t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
operator- ( const cuddcomplex a, const cuddcomplex b )
{
  cuddcomplex t;
  sub2 ( a.x, b.x, t.x,  a.y, b.y, t.y );
  return t;
}

__host__ __device__ __forceinline__ void
operator-= ( cuddcomplex &a, const cuddcomplex b )
{
  a = (a - b);
}

__host__ __device__ __forceinline__ cuddcomplex
operator* ( const cuddcomplex a, const cuddcomplex b )
{
  cuddcomplex t;
  cuddreal e = -a.y;
  mul2 ( a.x, b.x, t.x,  a.x, b.y, t.y );
  fma2 ( e,   b.y, t.x, t.x,  a.y, b.x, t.y, t.y );
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
operator* ( const cuddcomplex a, const cuddreal b )
{
  cuddcomplex t;
  mul2 ( a.x, b, t.x,  a.x, b, t.y );
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
operator* ( const cuddreal a, const cuddcomplex b )
{
  cuddcomplex t = (b * a);
  return t;
}

__host__ __device__ __forceinline__ void
operator*= ( cuddcomplex &a, const cuddcomplex b )
{
  a = (a * b);
}

__host__ __device__ __forceinline__ void
operator*= ( cuddcomplex &a, const cuddreal b )
{
  a = (a * b);
}

__host__ __device__ __forceinline__ cuddcomplex
fma ( const cuddcomplex a, const cuddcomplex b, const cuddcomplex c )
{
  cuddcomplex t;
  cuddreal e = -a.y;
  fma2 ( a.x, b.x, c.x, t.x,  a.x, b.y, c.y, t.y );
  fma2 ( e,   b.y, t.x, t.x,  a.y, b.x, t.y, t.y );
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
fma ( const cuddcomplex a, const cuddreal b, const cuddcomplex c )
{
  cuddcomplex t;
  fma2 ( a.x, b, c.x, t.x,  a.y, b, c.y, t.y );
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
fma ( const cuddreal a, const cuddcomplex b, const cuddcomplex c )
{
  cuddcomplex t = fma( b, a, c );
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
CUWDIV ( const cuddcomplex a, const cuddcomplex b )
{
  /*
   * almost similar approach to divide a complex number as cuComplex.h
   */
  const cuddreal ZERO_ = { (double)0, (double)0 };
  const cuddcomplex ZERO = { ZERO_, ZERO_ };
  if ( b == ZERO ) {
    const cuddreal NAN_DD = { NAN, NAN };
    const cuddcomplex NAN_DDC = { NAN_DD, NAN_DD };
    return NAN_DDC;
  }
  if ( a == ZERO ) { return ZERO; }
  const cuddreal ONE_ = { (double)1, (double)0 };
  const cuddcomplex ONE  = { ONE_, ZERO_ };
  if ( b == ONE  ) { return a; }
  const cuddreal MONE_ = { (double)1, (double)0 };
  const cuddcomplex MONE = { MONE_, ZERO_ };
  if ( b == MONE ) { return -a; }

  cuddreal s = Abs(b.x) + Abs(b.y);
  cuddreal r = ONE_ / s;
  const cuddreal ar = a.x * r;
  const cuddreal ai = a.y * r;
  const cuddreal br = b.x * r;
  const cuddreal bi = b.y * r;
  s = (br * br) + (bi * bi);
  r = ONE_ / s;
  const cuddcomplex t = {
    ((ar * br) + (ai * bi)) * r,
    ((ai * br) - (ar * bi)) * r };
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
operator/ ( const cuddcomplex a, const cuddcomplex b )
{
  const cuddcomplex t = CUWDIV ( a, b );
  return t;
}

__host__ __device__ __forceinline__ void
operator/= ( cuddcomplex &a, const cuddcomplex b )
{
  a = ( a / b );
}


__host__ __device__ __forceinline__ void
add2 ( const cuddcomplex a1, const cuddcomplex b1, cuddcomplex &d1,
       const cuddcomplex a2, const cuddcomplex b2, cuddcomplex &d2 )
{
  cuddcomplex t1, t2;
  add2 ( a1.x, b1.x, t1.x,  a2.x, b2.x, t2.x );
  add2 ( a1.y, b1.y, t1.y,  a2.y, b2.y, t2.y );
  d1 = t1; d2 = t2;
}

__host__ __device__ __forceinline__ void
sub2 ( const cuddcomplex a1, const cuddcomplex b1, cuddcomplex &d1,
       const cuddcomplex a2, const cuddcomplex b2, cuddcomplex &d2 )
{
  cuddcomplex t1, t2;
  sub2 ( a1.x, b1.x, t1.x,  a2.x, b2.x, t2.x );
  sub2 ( a1.y, b1.y, t1.y,  a2.y, b2.y, t2.y );
  d1 = t1; d2 = t2;
}

__host__ __device__ __forceinline__ void
mul2 ( const cuddcomplex a1, const cuddcomplex b1, cuddcomplex &d1,
       const cuddcomplex a2, const cuddcomplex b2, cuddcomplex &d2 )
{
  cuddcomplex t1, t2;
  cuddreal e1 = -a1.y, e2 = -a2.y;
  mul2 ( a1.x, b1.x, t1.x,  a2.x, b2.x, t2.x );
  mul2 ( a1.x, b1.y, t1.y,  a2.x, b2.y, t2.y );
  fma2 ( e1,   b1.y, t1.x, t1.x,  e2,   b1.y, t2.x, t2.x );
  fma2 ( a1.y, b1.x, t1.y, t1.y,  a2.y, b1.x, t2.y, t2.y );
  d1 = t1; d2 = t2;
}

__host__ __device__ __forceinline__ void
fma2 ( const cuddcomplex a1, const cuddcomplex b1, const cuddcomplex c1, cuddcomplex &d1,
       const cuddcomplex a2, const cuddcomplex b2, const cuddcomplex c2, cuddcomplex &d2 )
{
  cuddcomplex t1, t2;
  cuddreal e1 = -a1.y, e2 = -a2.y;
  fma2( a1.x, b1.x, c1.x, t1.x,  a2.x, b2.x, c2.x, t2.x );
  fma2( a1.x, b1.y, c1.y, t1.y,  a2.x, b2.y, c2.y, t2.y );
  fma2( e1,   b1.y, t1.x, t1.x,  e2,   b2.y, t2.x, t2.x );
  fma2( a1.y, b1.x, t1.y, t1.y,  a2.y, b2.x, t2.y, t2.y );
  d1 = t1; d2 = t2;
}


__host__ __device__ __forceinline__ cuddreal
Abs ( const cuddcomplex a )
{
  const cuddreal r2 = a.x*a.x + a.y*a.y;
  const cuddreal t = Sqrt( r2 );
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
Conj ( const cuddcomplex a )
{
  const cuddreal tx =  a.x;
  const cuddreal ty = -a.y;
  const cuddcomplex t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ cuddcomplex
__choose__ ( const bool flag, const cuddcomplex a, const cuddcomplex b )
{
  const cuddreal tx = __choose__( flag, a.x, b.x );
  const cuddreal ty = __choose__( flag, a.y, b.y );
  const cuddcomplex t = { tx, ty };
  return t;
}

#endif

#define ASPEN_DDCOMPLEX_H_INCLUDED		1
#endif

