#include "aspen_real.h"
#ifndef ASPEN_DDREAL_H_INCLUDED

struct __align__(16) __cuddreal__
{
  double x, y;
};

typedef	struct __cuddreal__		cuddreal;


#ifdef __cplusplus

#ifdef __CUDA_ARCH__
#if GPU_ARCH>300
__device__ __forceinline__ cuddreal
__ldg ( const cuddreal * a )
{
  const double tx = __ldg( &(a->x) );
  const double ty = __ldg( &(a->y) );
  const cuddreal t = { tx, ty };
  return t;
}
#endif
#endif


__host__ __device__ __forceinline__ bool
operator== ( const cuddreal a, const cuddreal b )
{
  const bool ret = ((a.x == b.x) && (a.y == b.y)) ? true : false;
  return ret;
}

__host__ __device__ __forceinline__ bool
operator!= ( const cuddreal a, const cuddreal b)
{
  const bool ret = ((a.x != b.x) || (a.y != b.y)) ? true : false;
  return ret;
}

__host__ __device__ __forceinline__ bool
operator< ( const cuddreal a, const cuddreal b)
{
  bool ret;
  if ( a.x == b.x ) {
    ret = (a.y < b.y) ? true : false;
  } else {
    ret = (a.x < b.x) ? true : false;
  }
  return ret;
}

__host__ __device__ __forceinline__ bool
operator> ( const cuddreal a, const cuddreal b)
{
  return ( b < a );
}

__host__ __device__ __forceinline__ bool
operator<= ( const cuddreal a, const cuddreal b)
{
  bool ret;
  if ( a.x == b.x ) {
    ret = (a.y <= b.y) ? true : false;
  } else {
    ret = (a.x < b.x) ? true : false;
  }
  return ret;
}

__host__ __device__ __forceinline__ bool
operator>= ( const cuddreal a, const cuddreal b)
{
  return ( b <= a );
}


__host__ __device__ __forceinline__ double
__fmad ( const double a,  const double b, const double c )
{
  return  fma ( a, b, c );
}


__host__ __device__ __forceinline__ void
cuWTwoSum ( const double a, const double b, double &s, double &e )
{
  const double  S = __add ( a, b );
  const double  u = __sub ( S, a );
  const double  v = __sub ( S, u );
  const double  p = __sub ( a, v );
  const double  q = __sub ( b, u );
  const double  E = __add ( p, q );
  s = S; e = E;
}

__host__ __device__ __forceinline__ void
cuWTwoSum2 ( const double a1, const double b1, double &s1, double &e1,
             const double a2, const double b2, double &s2, double &e2 )
{
  const double  S1 = __add ( a1, b1 );
  const double  S2 = __add ( a2, b2 );
  const double  u1 = __sub ( S1, a1 );
  const double  u2 = __sub ( S2, a2 );
  const double  v1 = __sub ( S1, u1 );
  const double  v2 = __sub ( S2, u2 );
  const double  p1 = __sub ( a1, v1 );
  const double  p2 = __sub ( a2, v2 );
  const double  q1 = __sub ( b1, u1 );
  const double  q2 = __sub ( b2, u2 );
  const double  E1 = __add ( p1, q1 );
  const double  E2 = __add ( p2, q2 );
  s1 = S1; e1 = E1;
  s2 = S2; e2 = E2;
}

__host__ __device__ __forceinline__ void
cuWQuickTwoSum ( const double a, const double b, double &s, double &e )
{
  const double  S = __add ( a, b );
  const double  v = __sub ( S, a );
  const double  E = __sub ( b, v );
  s = S; e = E;
}

__host__ __device__ __forceinline__ void
cuWQuickTwoSum2 ( const double a1, const double b1, double &s1, double &e1,
                  const double a2, const double b2, double &s2, double &e2 )
{
  const double  S1 = __add ( a1, b1 );
  const double  S2 = __add ( a2, b2 );
  const double  v1 = __sub ( S1, a1 );
  const double  v2 = __sub ( S2, a2 );
  const double  E1 = __sub ( b1, v1 );
  const double  E2 = __sub ( b2, v2 );
  s1 = S1; e1 = E1;
  s2 = S2; e2 = E2;
}


__host__ __device__ __forceinline__ cuddreal
CUWADD ( const cuddreal a, const cuddreal b )
{
  cuddreal t, c;
  cuWTwoSum ( a.x, b.x, t.x, t.y );
  const double  r = __add( a.y, b.y );
  t.y = __add ( t.y, r );
  cuWQuickTwoSum ( t.x, t.y, c.x, c.y );
  return c;
}

__host__ __device__ __forceinline__ void
CUWADD2 ( const cuddreal a1, const cuddreal b1, cuddreal &d1,
          const cuddreal a2, const cuddreal b2, cuddreal &d2 )
{
  cuddreal t1,t2, c1,c2;
  cuWTwoSum2 ( a1.x, b1.x, t1.x, t1.y,
               a2.x, b2.x, t2.x, t2.y );
  const double  r1 = __add( a1.y, b1.y );
  const double  r2 = __add( a2.y, b2.y );
  t1.y = __add ( t1.y, r1 );
  t2.y = __add ( t2.y, r2 );
  cuWQuickTwoSum2 ( t1.x, t1.y, c1.x, c1.y,
                    t2.x, t2.y, c2.x, c2.y );
  d1 = c1; d2 = c2;
}

__host__ __device__ __forceinline__ cuddreal
CUWSUB ( const cuddreal a, const cuddreal b )
{
  cuddreal t, c;
  cuWTwoSum ( a.x, -b.x, t.x, t.y );
  t.y = __add ( t.y, __sub ( a.y, b.y ) );
  cuWQuickTwoSum ( t.x, t.y, c.x, c.y );
  return c;
}

__host__ __device__ __forceinline__ void
CUWSUB2 ( const cuddreal a1, const cuddreal b1, cuddreal &d1,
          const cuddreal a2, const cuddreal b2, cuddreal &d2 )
{
  cuddreal t1,t2, c1,c2;
  cuWTwoSum ( a1.x, -b1.x, t1.x, t1.y );
  cuWTwoSum ( a2.x, -b2.x, t2.x, t2.y );
  t1.y = __add ( t1.y, __sub ( a1.y, b1.y ) );
  t2.y = __add ( t2.y, __sub ( a2.y, b2.y ) );
  cuWQuickTwoSum ( t1.x, t1.y, c1.x, c1.y );
  cuWQuickTwoSum ( t2.x, t2.y, c2.x, c2.y );
  d1 = c1; d2 = c2;
}

__host__ __device__ __forceinline__ cuddreal
CUWMUL ( const cuddreal a, const cuddreal b )
{
  double t, e;
  cuddreal c;
  t   = __fmad ( a.y, b.x, __mul( a.x, b.y ) );
  c.x = __fmad ( a.x, b.x, t );
  e   = __fmad ( a.x, b.x, -c.x );
  c.y = __add ( e, t );
  return c;
}

__host__ __device__ __forceinline__ cuddreal
CUSMUL ( const cuddreal a, const double b )
{
  double t, e;
  cuddreal c;
  t   = __mul ( a.y, b );
  c.x = __fmad ( a.x, b, t );
  e   = __fmad ( a.x, b, -c.x );
  c.y = __add ( e, t );
  return c;
}

__host__ __device__ __forceinline__ void
CUWMUL2 ( const cuddreal a1, const cuddreal b1, cuddreal &d1,
          const cuddreal a2, const cuddreal b2, cuddreal &d2 )
{
  double t1,t2, e1,e2;
  double s1,s2;
  cuddreal c1,c2;
  s1   = __mul( a1.x, b1.y );
  s2   = __mul( a2.x, b2.y );
  t1   = __fmad ( a1.y, b1.x, s1 );
  t2   = __fmad ( a2.y, b2.x, s2 );
  c1.x = __fmad ( a1.x, b1.x, t1 );
  c2.x = __fmad ( a2.x, b2.x, t2 );
  s1   = -c1.x;
  s2   = -c2.x;
  e1   = __fmad ( a1.x, b1.x, s1 );
  e2   = __fmad ( a2.x, b2.x, s2 );
  c1.y = __add ( e1, t1 );
  c2.y = __add ( e2, t2 );
  d1 = c1; d2 = c2;
}


__host__ __device__ __forceinline__ cuddreal
operator+ ( const cuddreal a )
{
  return a;
}

__host__ __device__ __forceinline__ cuddreal
operator+ ( const cuddreal a, const cuddreal b )
{
  return CUWADD ( a, b );
}

__host__ __device__ __forceinline__ void
operator+= ( cuddreal &a, const cuddreal b )
{
  a = CUWADD ( a, b );
}

__host__ __device__ __forceinline__ cuddreal
operator- ( const cuddreal a )
{
  const double tx = -a.x;
  const double ty = -a.y;
  const cuddreal  t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ cuddreal
operator- ( const cuddreal a, const cuddreal b )
{
  return CUWSUB ( a, b );
}

__host__ __device__ __forceinline__ void
operator-= ( cuddreal &a, const cuddreal b )
{
  a = CUWSUB ( a, b );
}

__host__ __device__ __forceinline__ cuddreal
operator* ( const cuddreal a, const cuddreal b )
{
  return CUWMUL ( a, b );
}

__host__ __device__ __forceinline__ cuddreal
operator* ( const cuddreal a, const double b )
{
  return CUSMUL ( a, b );
}

__host__ __device__ __forceinline__ cuddreal
operator* ( const double a, const cuddreal b )
{
  return CUSMUL ( b, a );
}

__host__ __device__ __forceinline__ void
operator*= ( cuddreal &a, const cuddreal b )
{
  a = ( a * b );
}

__host__ __device__ __forceinline__ void
operator*= ( cuddreal &a, const double b )
{
  a = ( a * b );
}

__host__ __device__ __forceinline__ cuddreal
fma ( const cuddreal a, const cuddreal b, const cuddreal c )
{
  cuddreal  t = (a * b) + c;
  return t;
}

__host__ __device__ __forceinline__ cuddreal
fma ( const cuddreal a, const double b, const cuddreal c )
{
  cuddreal  t = (a * b) + c;
  return t;
}

__host__ __device__ __forceinline__ cuddreal
fma ( const double a, const cuddreal b, const cuddreal c )
{
  cuddreal  t = (a * b) + c;
  return t;
}

__host__ __device__ __forceinline__ cuddreal
CUWDIV ( const cuddreal a, const cuddreal b )
{
  const cuddreal ZERO = { (double)0, (double)0 };
  if ( b == ZERO ) {
    const cuddreal NAN_DD = { NAN, NAN };
    return NAN_DD;
  }
  if ( a == ZERO ) { return ZERO; }
  const cuddreal ONE  = { (double)1, (double)0 };
  if ( b == ONE  ) { return a; }
  const cuddreal MONE = { (double)-1, (double)0 };
  if ( b == MONE ) { return -a; }

  /* A.H.Karp and P.Markstein:
   * High Precision Division and Square Root,
   * ATOMS, 23(4), 1997, p.561-589.
   *
   *  when x=Approx(a/b),
   *  y = b*x
   *  Y = y + x(b-a*y) = goodApprox(a/b)
   */
  const double   xn = a.x / b.x;
  const cuddreal yn = b * xn;
  const cuddreal T  = a * yn;
  cuddreal       t  = b - T;
  t *= xn;
  const cuddreal Y  = yn + t;

  return Y;
}

__host__ __device__ __forceinline__ cuddreal
operator/ ( const cuddreal a, const cuddreal b )
{
  const cuddreal t = CUWDIV ( a, b );
  return t;
}

__host__ __device__ __forceinline__ void
operator/= ( cuddreal &a, const cuddreal b )
{
  a = ( a / b );
}


__host__ __device__ __forceinline__ void
add2 ( const cuddreal a1, const cuddreal b1, cuddreal &d1,
       const cuddreal a2, const cuddreal b2, cuddreal &d2 )
{
  cuddreal  t1,t2;
  CUWADD2 ( a1, b1, t1, a2, b2, t2 );
  d1 = t1; d2 = t2;
}

__host__ __device__ __forceinline__ void
sub2 ( const cuddreal a1, const cuddreal b1, cuddreal &d1,
       const cuddreal a2, const cuddreal b2, cuddreal &d2 )
{
  cuddreal  t1,t2;
  CUWSUB2 ( a1, b1, t1, a2, b2, t2 );
  d1 = t1; d2 = t2;
}

__host__ __device__ __forceinline__ void
mul2 ( const cuddreal a1, const cuddreal b1, cuddreal &d1,
       const cuddreal a2, const cuddreal b2, cuddreal &d2 )
{
  cuddreal  t1,t2;
  CUWMUL2 ( a1, b1, t1, a2, b2, t2 );
  d1 = t1; d2 = t2;
}

__host__ __device__ __forceinline__ void
fma2 ( const cuddreal a1, const cuddreal b1, const cuddreal c1, cuddreal &d1,
       const cuddreal a2, const cuddreal b2, const cuddreal c2, cuddreal &d2 )
{
  cuddreal  t1,t2;
  CUWMUL2 ( a1, b1, t1, a2, b2, t2 );
  CUWADD2 ( c1, t1, t1, c2, t2, t2 );
  d1 = t1; d2 = t2;
}


__host__ __device__ __forceinline__ cuddreal
Abs ( const cuddreal a )
{
  const double tx = fabs(a.x);
  const double ty = fabs(a.y);
  const cuddreal  t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ cuddreal
Conj ( const cuddreal a )
{
  return a;
}


__host__ __device__ __forceinline__ cuddreal
Sqrt ( const cuddreal a )
{
  const cuddreal ZERO = { (double)0, (double)0 };
  if ( a == ZERO ) { return a; };
  if ( a < ZERO ) {
    const cuddreal NAN_DD = { NAN, NAN };
    return NAN_DD;
  }

  /* A.H.Karp and P.Markstein:
   * High Precision Division and Square Root,
   * ATOMS, 23(4), 1997, p.561-589.
   *
   *  when x=Approx(1/sqrt(a)),
   *  y = a*x
   *  Y = y + (x/2)(a-y^2) = goodApprox(sqrt(a))
   */
  double         xn = 1e0 / sqrt(a.x);
  const cuddreal yn = a * xn;
  const cuddreal T  = yn * yn;
  cuddreal       t  = a - T;
  xn /= 2; t *= xn;
  const cuddreal Y  = yn + t;

  return Y;
}

__host__ __device__ __forceinline__ cuddreal
__choose__ ( const bool flag, const cuddreal a, const cuddreal b )
{
  const double  tx = (flag ? a.x : b.x);
  const double  ty = (flag ? a.y : b.y);
  const cuddreal  t = { tx, ty };
  return t;
}

#endif

#define ASPEN_DDREAL_H_INCLUDED		1
#endif

